from betterbot import Setup, Intents, Events
from example2depend import Ban
import asyncio
import json
from betterbot import CommandTypes

class Bot:
    def __init__(self):
        with open("config.json") as f:
            self.config = json.load(f)
        
        token = self.config["token"]
        mod_roles = self.config["mod_roles"]
        ban_role = self.config["ban_role"]

        self.guild = self.config["guild"]    

        self.bot = Setup(token=token, intents=Intents.all)
        self.bot.debug_logs = False
        self.event = self.bot.event
        self.rest_client = self.bot.rest_client
        self.ban = Ban(mod_roles, self.rest_client, self.guild, ban_role)
        self.tasks = set() # Needed to avoid GC

    async def start(self):
        await self.event.register(Events.BOT_READY, self.on_ready)
        await self.bot.start()

    async def on_ready(self, username, application_id):
        options = [{
            "name": "user", 
            "type": CommandTypes.USER, 
            "description": "User to ban",
            "required": True,
            },
            {
                "name": "unit",
                "description": "Unit of time",
                "type": CommandTypes.STRING,
                "required": True,
                "choices": [
                    {"name": "Minutes", "value": "minutes"},
                    {"name": "Hours", "value": "hours"},
                    {"name": "Days", "value": "days"},
                ]
            },
            {
                "name": "duration",
                "description": "Amount",
                "type": CommandTypes.INTEGER,
                "required": True,
            },
            ]
        
        await self.rest_client.create_guild_command(self.ban.ban, application_id=application_id, guild_id=self.guild, name="ban", description="Ban a user", options=options)
    
        options = [{
            "name": "user",  
            "description": "User to unban",
            "type": CommandTypes.USER,
            "required": True
        }]

        await self.rest_client.create_guild_command(self.ban.unban, application_id=application_id, guild_id=self.guild, name="unban", description="Unban someone", options=options)

        task = asyncio.create_task(self.ban.polling_unban())
        self.tasks.add(task)
        task.add_done_callback(self.tasks.discard)

        print(f"Logged in as {username}")

asyncio.run(Bot().start())


import json
import asyncio
import logging
from .. import Setup
from ..gateway_client.command_related.commands import Commands

class Test:
    def __init__(self):
        self.setup = Setup(token="I like stuff", intents="I do too")
        self.logger = logging.getLogger(__name__)
        self.setup.debug_logs = True
        self.setup.test_mode = True
        self.commandbus_flag = False
        self.eventbus_flag = False
        self.times = 0
        self.success_count = 0

    async def mock_command_receiver(self, **kwargs):
        if not kwargs:
            self.logger.info(f"mock_command_receiver received nothing. 41298491")
        else:
            self.logger.info(f"mock_command_receiver received {kwargs}. 41298491")

    async def mock_eventbus_receiver(self, times):
        self.times += 1
        print(f"You should see me {times}")
        return 69

    async def debug_commandbus(self):
        try:
            await self.commandbus.emit("text")
        except RuntimeError:
            self.logger.info("Commandbus successfully threw error when there is nothing registered 34872783")
            self.success_count += 1

        try:
            await self.commandbus.register("test", lambda: self.mock_eventbus_receiver(times=2))
            await self.commandbus.register("test", lambda: self.mock_eventbus_receiver(times=2))
        except RuntimeError:
            self.logger.info("Commandbus correctly raised exception. 93242873")
            self.success_count += 1

        finally:
            nothing = await self.commandbus.emit("test", fire_and_forget=True)
            something = await self.commandbus.emit("test", fire_and_forget=False)
            await asyncio.sleep(1)

            assert something == 69, "Return values not expected"

        self.logger.warning("COMMANDBUS DEBUGGING FINISHED")

    async def debug_heartbeat_service(self):
        self.session.sequence = 5
        self.session.heartbeat_interval = 2

        response = await self.heartbeat_service._start_heartbeat_watchdog()
    
        assert response == "WATCHDOG_TRIGGERED", "Watchdog not triggered."
        self.success_count += 1

        response = await self.heartbeat_service.send_heartbeat()
        assert response == "SENT", "Heartbeat not working as expected"
        self.success_count += 1
        
        self.logger.warning("HEARTBEAT SERVICE DEBUGGING FINISHED 12483251")

    async def debug_eventbus(self):
        asyncio.create_task(self.eventbus.wait_for_event("test696969"))

        await self.eventbus.register("test696969", lambda: self.mock_eventbus_receiver(times=3), lambda: self.mock_eventbus_receiver(times=3))
        await self.eventbus.register("test696969", lambda: self.mock_eventbus_receiver(times=3))

        await self.eventbus.emit("test696969")

        self.logger.warning("EVENTBUS DEBUGGING FINISHED 01958235")

    async def debug_reset_websocket_handler(self): # Error prone therefore it has it's own function
        self.websocket_handler.reset_running = True

        message = await self.websocket_handler.reset()

        assert message == "BUSY", "The websocket reset function continued running even tho flag is set. 103248124"

        self.websocket_handler.reset_running = False
        self.websocket_handler.gateway_url = "wss://ws.ifelse.io"

        await self.websocket_handler.reset()

        await self.websocket_handler.ws.close()

        self.websocket_handler.gateway_url = "I am invalid"

        task = asyncio.create_task(self.websocket_handler.reset())

        await asyncio.sleep(10)
        self.logger.info("You should have seen some reconnect attempts failing. 29913189")
        task.cancel()

        self.websocket_handler.gateway_url = "wss://ws.ifelse.io"
        task = asyncio.create_task(self.websocket_handler.send_websocket_message(self)) # Will throw error and try to reset.
        await asyncio.sleep(10)
        task.cancel()
        self.logger.info("You should have seen some reconnect attempts working. 1284184")

        self.logger.warning("WEBSOCKET RESET HANDLER DEBUGGING FINISHED 1942925")

    async def debug_websocket_handler(self):
        self.websocket_handler.ws = self

        await self.websocket_handler.send_websocket_message("I like beautiful women.")
        await self.websocket_handler.receive_message()

        self.logger.warning("WEBSOCKET HANDLER DEBUGGING FINISHED 42356463")

    async def send(self, data):
        print(f"I like cool women. What about you? {data} 10994771")

    async def recv(self):
        return json.dumps("I hate cool women.")
    
    async def debug(self, debug_all=True):
        result = await self.setup.start()

        (self.commandbus, self.session, self.identify, self.heartbeat_service,
        self.handle_responses, self.websocket_handler, self.eventbus) = result

        await self.commandbus.register(Commands.STOP_HB_WATCHDOG, self.mock_command_receiver)
        await self.commandbus.register(Commands.STOP_HB_LOOP, self.mock_command_receiver)
        await self.commandbus.register(Commands.STOP_LISTENER, self.mock_command_receiver)
        await self.commandbus.register(Commands.SEND_WEBSOCKET_MESSAGE, self.mock_command_receiver)
        await self.commandbus.register(Commands.START_LISTENER, self.mock_command_receiver)
        await self.commandbus.register(Commands.CLIENT_RESET_NEEDED, self.websocket_handler.reset)

        if debug_all is True:
            await self.debug_websocket_handler()
            await self.debug_reset_websocket_handler()
            await self.debug_eventbus()
            await self.debug_commandbus()
            await self.debug_heartbeat_service()
            assert self.times == 5, "Times receiver ran not matching expected"
            assert self.success_count == 4
        else:
            return self.debug_websocket_handler, self.debug_reset_websocket_handler, self.debug_eventbus, self.debug_commandbus, self.debug_heartbeat_service

debug = asyncio.run(Test().debug(debug_all=True))
#asyncio.run(debug[2]())



from ..rest_client.wrapper import HttpWrapper
import logging
import asyncio
import sys

class Test:
    def __init__(self):
        self.tasks = set()
        self.spam_mode = False
        
        token = ""

        sys.stdout.reconfigure(encoding="utf-8")

        logging.basicConfig(
            level=logging.DEBUG,
            handlers=[
                logging.FileHandler("log.txt", mode="w", encoding="utf-8"),
                logging.StreamHandler(sys.stdout)
            ],
        )

        self.wrapper = HttpWrapper(token=token)

    async def debug_wrapper(self):
        channel_ids = [1482287900585496706, 1482287927206740008]
        json = {"content": "I am nice jk."}
        method = "post"
        if self.spam_mode is True:
            for l in range(100):
                for i in channel_ids:
                    endpoint = f"https://discord.com/api/v9/channels/{i}/messages"
                    task = asyncio.create_task(self.wrapper.request(method=method, endpoint=endpoint, identifier=f"{method}:{i}", json=json))
                    self.tasks.add(task)
                    task.add_done_callback(self.tasks.discard)
        else:
            for l in range(100):
                for i in channel_ids:
                    endpoint = f"https://discord.com/api/v9/channels/{i}/messages"
                    result = await self.wrapper.request(method=method, endpoint=endpoint, identifier=f"{method}:{i}", json=json)
                    print(result)

        await asyncio.Future()
asyncio.run(Test().debug_wrapper())


import asyncio
import logging

class Commandbus:
    def __init__(self):
        self.registry = {}
        self.tasks = set()
        self.logger = logging.getLogger(__name__)

    async def register(self, command_type, handler):
        if command_type in self.registry:
            raise RuntimeError("Only one handler per command allowed!")

        self.registry.update({command_type: handler})

        self.logger.debug(f"Registered command {command_type} to {handler}")

    async def emit(self, command_type, fire_and_forget=False, **kwargs):
        if command_type not in self.registry:
            raise RuntimeError("Undefined command was called!")

        handler = self.registry[command_type]

        self.logger.debug(f"Emitting {command_type} to {handler}")

        if fire_and_forget is False:
            return await handler(**kwargs)
        else:
            task = asyncio.create_task(handler(**kwargs))
            self.tasks.add(task)
            task.add_done_callback(self.tasks.discard)


from .endpoints import Endpoints
from ...gateway_client.event_related.events import Events
from ...enums import ResponseTypes

class Application:
    def __init__(self, wrapper, eventbus):
        self.send = wrapper.request
        self.register = eventbus.register

    # -------------------------
    # GLOBAL COMMANDS
    # -------------------------

    async def get_global_commands(self, application_id):
        endpoint = Endpoints.APPLICATION_GLOBAL_COMMANDS(d={"application_id": application_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"get_global_commands:{application_id}")

    async def create_global_command(self, application_id, name, description, options=None, cmd_type=1, *handlers):
        payload = {"name": name, "description": description, "type": cmd_type}
        if options: payload["options"] = options
        endpoint = Endpoints.APPLICATION_GLOBAL_COMMANDS(d={"application_id": application_id})
        response = await self.send(method="post", endpoint=endpoint, identifier=f"create_global_command:{application_id}", json=payload)
        
        if response != ResponseTypes.FAILED and response != ResponseTypes.CONNECTION_FAILED:
            event = f"{Events.INTERACTION_CREATE}:{name}"
            await self.register(event, *handlers)
        
        return response
    
    async def get_global_command(self, application_id, command_id):
        endpoint = Endpoints.APPLICATION_COMMAND(d={"application_id": application_id, "command_id": command_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"get_global_command:{application_id}")

    async def update_global_command(self, application_id, command_id, name=None, description=None, options=None):
        payload = {}
        if name: payload["name"] = name
        if description: payload["description"] = description
        if options: payload["options"] = options
        endpoint = Endpoints.APPLICATION_COMMAND(d={"application_id": application_id, "command_id": command_id})
        return await self.send(method="patch", endpoint=endpoint, identifier=f"update_global_command:{application_id}", json=payload)

    async def delete_global_command(self, application_id, command_id):
        endpoint = Endpoints.APPLICATION_COMMAND(d={"application_id": application_id, "command_id": command_id})
        return await self.send(method="delete", endpoint=endpoint, identifier=f"delete_global_command:{application_id}")

    # -------------------------
    # GUILD COMMANDS
    # -------------------------

    async def get_guild_commands(self, application_id, guild_id):
        endpoint = Endpoints.APPLICATION_GUILD_COMMANDS(d={"application_id": application_id, "guild_id": guild_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"get_guild_commands:{application_id}:{guild_id}")

    async def create_guild_command(self, application_id, guild_id, name, description, options=None, cmd_type=1, *handlers):
        payload = {"name": name, "description": description, "type": cmd_type}
        if options: payload["options"] = options
        endpoint = Endpoints.APPLICATION_GUILD_COMMANDS(d={"application_id": application_id, "guild_id": guild_id})
        response = await self.send(method="post", endpoint=endpoint, identifier=f"create_guild_command:{application_id}:{guild_id}", json=payload)

        if response != ResponseTypes.FAILED and response != ResponseTypes.CONNECTION_FAILED:
            event = f"{Events.INTERACTION_CREATE}:{name}:{guild_id}"
            await self.register(event, *handlers)
        
        return response

    async def get_guild_command(self, application_id, guild_id, command_id):
        endpoint = Endpoints.APPLICATION_GUILD_COMMAND(d={"application_id": application_id, "guild_id": guild_id, "command_id": command_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"get_guild_command:{application_id}:{guild_id}")

    async def update_guild_command(self, application_id, guild_id, command_id, name=None, description=None, options=None):
        payload = {}
        if name: payload["name"] = name
        if description: payload["description"] = description
        if options: payload["options"] = options
        endpoint = Endpoints.APPLICATION_GUILD_COMMAND(d={"application_id": application_id, "guild_id": guild_id, "command_id": command_id})
        return await self.send(method="patch", endpoint=endpoint, identifier=f"update_guild_command:{application_id}:{guild_id}", json=payload)

    async def delete_guild_command(self, application_id, guild_id, command_id):
        endpoint = Endpoints.APPLICATION_GUILD_COMMAND(d={"application_id": application_id, "guild_id": guild_id, "command_id": command_id})
        return await self.send(method="delete", endpoint=endpoint, identifier=f"delete_guild_command:{application_id}:{guild_id}")

    # -------------------------
    # COMMAND PERMISSIONS
    # -------------------------

    async def get_command_permissions(self, application_id, guild_id, command_id):
        endpoint = Endpoints.APPLICATION_COMMAND_PERMISSIONS(d={"application_id": application_id, "guild_id": guild_id, "command_id": command_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"get_command_permissions:{application_id}:{guild_id}")

    async def update_command_permissions(self, application_id, guild_id, command_id, permissions):
        payload = {"permissions": permissions}
        endpoint = Endpoints.APPLICATION_COMMAND_PERMISSIONS(d={"application_id": application_id, "guild_id": guild_id, "command_id": command_id})
        return await self.send(method="put", endpoint=endpoint, identifier=f"update_command_permissions:{application_id}:{guild_id}", json=payload)


from typing import Any, List, Optional, Callable
from .rest_handlers.interaction import Interaction
from .rest_handlers.guild import Guild
from .rest_handlers.channel import Channel
from .wrapper import HttpWrapper
from .rest_handlers.application import Application
from .eventsend import es

class RESTClient:
    def __init__(self, token, eventbus):
        wrapper = HttpWrapper(token)

        self.channel = Channel(wrapper)
        self.guild = Guild(wrapper)
        self.interaction = Interaction(wrapper)
        self.application = Application(wrapper, eventbus)
                
    # -------------------------
    # CHANNEL METHODS
    # -------------------------

    async def send_message(self, channel_id: str, content: Optional[str] = None, embed: Optional[list] = None, components: Optional[list] = None) -> Any:
        """
        Send a message to a channel.
        
        Args:
            channel_id: ID of the channel.
            content: Text content of the message.
            embed: Optional embed dictionary.
            components: Buttons and such.
        """
        return await self.channel.send_message(channel_id, content, embed, components)

    async def get_messages(self, channel_id: str, limit: int = 100, before_id: Optional[str] = None) -> Any:
        """Retrieve messages from a channel, optionally before a specific message ID."""
        return await self.channel.get_messages(channel_id, limit, before_id)

    async def get_message(self, channel_id: str, message_id: str):
        return await self.channel.get_message(channel_id, message_id)

    async def delete_message(self, channel_id: str, message_id: str) -> Any:
        """Delete a specific message in a channel."""
        return await self.channel.delete_message(channel_id, message_id)

    async def add_reaction(self, channel_id: str, message_id: str, emoji: str, user_id: Optional[str] = None) -> Any:
        """Add a reaction to a message."""
        return await self.channel.add_reaction(channel_id, message_id, emoji, user_id)

    async def remove_reaction(self, channel_id: str, message_id: str, emoji: str, user_id: Optional[str] = None) -> Any:
        """Remove a reaction from a message, optionally for a specific user."""
        return await self.channel.remove_reaction(channel_id, message_id, emoji, user_id)

    async def bulk_delete_messages(self, channel_id: str, message_ids: List[str]) -> Any:
        """Bulk delete multiple messages from a channel."""
        return await self.channel.bulk_delete_messages(channel_id, message_ids)

    async def get_pins(self, channel_id: str) -> Any:
        """Get all pinned messages in a channel."""
        return await self.channel.get_pins(channel_id)

    async def pin_message(self, channel_id: str, message_id: str) -> Any:
        """Pin a message in a channel."""
        return await self.channel.pin_message(channel_id, message_id)

    async def unpin_message(self, channel_id: str, message_id: str) -> Any:
        """Unpin a message in a channel."""
        return await self.channel.unpin_message(channel_id, message_id)

    async def typing(self, channel_id: str) -> Any:
        """Trigger typing indicator in a channel."""
        return await self.channel.typing(channel_id)

    # -------------------------
    # GUILD METHODS
    # -------------------------

    async def get_guild_info(self, guild_id: str) -> Any:
        """Retrieve information about a guild."""
        return await self.guild.get_guild_info(guild_id)

    async def get_guild_members(self, guild_id: str, limit: int = 100, after: Optional[str] = None) -> Any:
        """Retrieve a list of guild members."""
        return await self.guild.get_guild_members(guild_id, limit, after)

    async def get_guild_member(self, guild_id: str, user_id: str) -> Any:
        """Retrieve information about a specific guild member."""
        return await self.guild.get_guild_member(guild_id, user_id)

    async def get_guild_roles(self, guild_id: str) -> Any:
        """Get all roles in a guild."""
        return await self.guild.get_guild_roles(guild_id)

    async def get_guild_role(self, guild_id: str, role_id: str) -> Any:
        """Retrieve information about a specific role."""
        return await self.guild.get_guild_role(guild_id, role_id)

    async def add_member_role(self, guild_id: str, user_id: str, role_id: str) -> Any:
        """Add a role to a guild member."""
        return await self.guild.add_member_role(guild_id, user_id, role_id)

    async def remove_member_role(self, guild_id: str, user_id: str, role_id: str) -> Any:
        """Remove a role from a guild member."""
        return await self.guild.remove_member_role(guild_id, user_id, role_id)

    async def member_roles(self, guild_id: str, user_id: str, roles: Optional[list] = None) -> Any:
        """Remove a or all roles from a guild member.

        Roles must be in a list if none all roles will be removed.
        
        This command is whitelist mode meaning all roles not specified will be removed.

        If you specify new roles it will add them but you must also specify current roles.

        """
        return await self.guild.member_roles(guild_id, user_id, roles)

    async def get_guild_bans(self, guild_id: str) -> Any:
        """Get a list of banned users in a guild."""
        return await self.guild.get_guild_bans(guild_id)

    async def get_guild_ban(self, guild_id: str, user_id: str) -> Any:
        """Get information about a specific guild ban."""
        return await self.guild.get_guild_ban(guild_id, user_id)

    async def prune_guild_members(self, guild_id: str, days: Optional[int] = None, compute_prune_count: bool = False) -> Any:
        """Prune inactive members from a guild."""
        return await self.guild.prune_guild_members(guild_id, days, compute_prune_count)

    async def get_guild_channels(self, guild_id: str) -> Any:
        """Get all channels in a guild."""
        return await self.guild.get_guild_channels(guild_id)

    # -------------------------
    # INTERACTION METHODS
    # -------------------------

    async def send_interaction_callback(self, interaction_id: str, interaction_token: str, cmd_type: int, content: Optional[str] = None, embed: Optional[list] = None, components: Optional[list] = None, ephemeral: bool = False) -> Any:
        """Send an interaction callback response."""
        return await self.interaction.send_interaction_callback(interaction_id, interaction_token, cmd_type, content, embed, components, ephemeral)

    async def send_interaction_followup(self, application_id: str, interaction_token: str, content: Optional[str] = None, embed: Optional[list] = None, components: Optional[list] = None, ephemeral: bool = False) -> Any:
        """Send a follow-up message for an interaction."""
        return await self.interaction.send_interaction_followup(application_id, interaction_token, content, embed, components, ephemeral)

    async def get_original_interaction_message(self, application_id: str, interaction_token: str) -> Any:
        """Retrieve the original interaction message."""
        return await self.interaction.get_original_interaction_message(application_id, interaction_token)

    async def update_interaction_message(self, application_id: str, interaction_token: str, content: Optional[str] = None, embed: Optional[list] = None, components: Optional[list] = None, message_id: Optional[str] = None) -> Any:
        """Update an interaction message, optionally specifying message ID."""
        return await self.interaction.update_interaction_message(application_id, interaction_token, content, embed, components, message_id)

    # -------------------------
    # APPLICATION COMMANDS
    # -------------------------

    async def get_global_commands(self, application_id: str) -> Any:
        """List all global application commands."""
        return await self.application.get_global_commands(application_id)

    async def create_global_command(self, *handlers: Callable, application_id: str, name: str, description: str, options: Optional[list] = None, cmd_type: Optional[int] = 1) -> Any:
        """Create a new global application command."""
        return await self.application.create_global_command(application_id, name, description, options, cmd_type, *handlers)

    async def get_global_command(self, application_id: str, command_id: str) -> Any:
        """Get a specific global application command by ID."""
        return await self.application.get_global_command(application_id, command_id)

    async def update_global_command(self, application_id: str, command_id: str, name: Optional[str] = None, description: Optional[str] = None, options: Optional[list] = None) -> Any:
        """Update a global application command."""
        return await self.application.update_global_command(application_id, command_id, name, description, options)

    async def delete_global_command(self, application_id: str, command_id: str) -> Any:
        """Delete a global application command."""
        return await self.application.delete_global_command(application_id, command_id)

    async def get_guild_commands(self, application_id: str, guild_id: str) -> Any:
        """List all guild-specific commands for an application."""
        return await self.application.get_guild_commands(application_id, guild_id)

    async def create_guild_command(self, *handlers: Callable, application_id: str, guild_id: str, name: str, description: str, options: Optional[list] = None, cmd_type: Optional[int] = 1) -> Any:
        """Create a new guild-specific application command."""
        return await self.application.create_guild_command(application_id, guild_id, name, description, options, cmd_type, *handlers)

    async def get_guild_command(self, application_id: str, guild_id: str, command_id: str) -> Any:
        """Get a specific guild command by ID."""
        return await self.application.get_guild_command(application_id, guild_id, command_id)

    async def update_guild_command(self, application_id: str, guild_id: str, command_id: str, name: Optional[str] = None, description: Optional[str] = None, options: Optional[list] = None) -> Any:
        """Update a guild-specific command."""
        return await self.application.update_guild_command(application_id, guild_id, command_id, name, description, options)

    async def delete_guild_command(self, application_id: str, guild_id: str, command_id: str) -> Any:
        """Delete a guild-specific command."""
        return await self.application.delete_guild_command(application_id, guild_id, command_id)

    async def get_command_permissions(self, application_id: str, guild_id: str, command_id: str) -> Any:
        """Get the permissions for a specific guild command."""
        return await self.application.get_command_permissions(application_id, guild_id, command_id)

    async def update_command_permissions(self, application_id: str, guild_id: str, command_id: str, permissions: list) -> Any:
        """Update the permissions for a specific guild command."""
        return await self.application.update_command_permissions(application_id, guild_id, command_id, permissions)


from typing import Optional, Any, Callable
from .rest_handlers.interaction import InteractionTypes

class es:
    def __init__(self, rest_client, **kwargs):
        self.rest_client = rest_client
        
        self.application_id = None
        self.channel = None
        self.interaction_id = None
        self.interaction_token = None

        for key, value in kwargs.items():
            setattr(self, key, value)

    async def create_global_command(self, *handlers: Callable, name: str, description: str, options: Optional[list] = None, cmd_type: int = 1) -> Any:
        """Create a new global application command."""
        if self.application_id is None:
            raise RuntimeError("You cannot register a command via Event Send (es) when there isn't an event with application ID. You should use this command in bot ready event.")
        
        return await self.rest_client.create_global_command(*handlers, application_id=self.application_id, name=name, description=description, options=options, cmd_type=cmd_type)

    async def create_guild_command(self, *handlers: Callable, guild_id: str, name: str, description: str, options: Optional[list] = None, cmd_type: int = 1) -> Any:
        """Create a new guild-specific application command."""

        if self.application_id is None:
            raise RuntimeError("You cannot register a command via Event Send (es) when there isn't an event with application ID. You should use this command in bot ready event.")
        
        return await self.rest_client.create_guild_command(*handlers, application_id=self.application_id, guild_id=guild_id, name=name, description=description, options=options, cmd_type=cmd_type)

    async def send(self, content: Optional[str] = None, embed: Optional[list] = None, components: Optional[list] = None) -> Any:
        """
        Send a message to a channel.
        
        Args:
            content: Text content of the message.
            embed: Optional embed dictionary.
            components: Buttons and such.
        """
        if self.channel is None:
            raise RuntimeError("You cannot send a message via Event Send (es) when there isn't an event with channel information.")
        
        return await self.rest_client.send_message(self.channel, content, embed, components)
    
    async def respond_message(self, content: Optional[str] = None, embed: Optional[list] = None, components: Optional[list] = None, ephemeral: bool = False) -> Any:
        """Send an interaction message response."""

        if self.interaction_id is None or self.interaction_token is None:
            raise RuntimeError("You cannot send a message via Event Send (es) when there isn't a command invoker.")

        return await self.rest_client.send_interaction_callback(self.interaction_id, self.interaction_token, InteractionTypes.RESPOND_WITH_MESSAGE, content, embed, components, ephemeral)

    async def send_interaction_callback(self, cmd_type: int, content: Optional[str] = None, embed: Optional[list] = None, components: Optional[list] = None, ephemeral: bool = False) -> Any:
        """Send an interaction response."""

        if self.interaction_id is None or self.interaction_token is None:
            raise RuntimeError("You cannot send an interaction callback via Event Send (es) when there isn't a command invoker.")

        return await self.rest_client.send_interaction_callback(self.interaction_id, self.interaction_token, cmd_type, content, embed, components, ephemeral)

    async def send_interaction_followup(self, content: Optional[str] = None, embed: Optional[list] = None, components: Optional[list] = None, ephemeral: bool = False) -> Any:
        """Send an interaction follow-up."""

        if self.application_id is None or self.interaction_token is None:
            raise RuntimeError("You cannot send a follow-up via Event Send (es) when there isn't a command invoker.")

        return await self.rest_client.send_interaction_followup(self.application_id, self.interaction_token, content, embed, components, ephemeral)
    
    async def get_original_interaction_message(self):
        """Retrieve the original interaction message."""
        
        if self.application_id is None or self.interaction_token is None:
            raise RuntimeError("You cannot send a get interaction message via Event Send (es) when there isn't a command invoker.")

        return await self.rest_client.get_original_interaction_message(self.application_id, self.interaction_token)
    
    async def update_interaction_message(self, content: Optional[str] = None, embed: Optional[list] = None, components: Optional[list] = None, message_id: Optional[str] = None) -> Any:
        """Update an interaction message, optionally specifying message ID."""

        if self.application_id is None or self.interaction_token is None:
            raise RuntimeError("You cannot update a interactiong message via Event Send (es) when there isn't a command invoker.")


        return await self.rest_client.update_interaction_message(self.application_id, self.interaction_token, content, embed, components, message_id)


import logging
from ..command_related.commands import Commands
import asyncio

class HeartbeatService:
    def __init__(self, commandbus, session):
        self.logger = logging.getLogger(__name__)

        self.emit = commandbus.emit
        self.session = session
        self.watchdog = None
        self.heartbeat_loop = None
        self.payload = {"op": 1, "d": session.sequence}

    async def _start_heartbeat_watchdog(self):
        try:
            self.logger.debug("Started heartbeat watchdog.")
            await asyncio.sleep(self.session.heartbeat_interval * 1.5)
            self.logger.warning("I AM BARKING I AM BARKING! DOG ALERT. Watchdog triggered. Discord didn't respond with Heartbeat ACK.")
            await self.emit(Commands.CLIENT_RESET_NEEDED, fire_and_forget=True) # Fire amd forget needed for safety
            return "WATCHDOG_TRIGGERED"

        except asyncio.CancelledError: pass

    async def _loop_send_heartbeat(self):
        self.logger.debug(f"Heartbeat loop started with heartbeat interval of {self.session.heartbeat_interval}")
        while True:
            await asyncio.sleep(self.session.heartbeat_interval)
            asyncio.create_task(self.send_heartbeat())

    async def stop_hb_watchdog(self):
        if isinstance(self.watchdog, asyncio.Task) and not self.watchdog.done():
            self.watchdog.cancel()
            self.logger.debug("Heartbeat watchdog stopped.")

    async def stop_hb_loop(self):
        if isinstance(self.heartbeat_loop, asyncio.Task) and not self.heartbeat_loop.done():
            self.heartbeat_loop.cancel()
            self.logger.debug("Heartbeat loop stopped.")

    async def start_hb_watchdog(self):            
        await self.stop_hb_watchdog()
        self.watchdog = asyncio.create_task(self._start_heartbeat_watchdog())

    async def start_hb_loop(self):
        await self.stop_hb_loop()
        self.heartbeat_loop = asyncio.create_task(self._loop_send_heartbeat())

    async def send_heartbeat(self):
        await self.emit(Commands.SEND_WEBSOCKET_MESSAGE, fire_and_forget=True, data=self.payload)
        self.logger.debug("Sent a heartbeat")
        return "SENT"


import logging
import asyncio
from ..command_related.commands import Commands

class HandleResponses:
    def __init__(self, commandbus, session):
        self.emit = commandbus.emit
        self.session = session
        self.logger = logging.getLogger(__name__)
        self.listener = None

    async def _listener(self):
        try:
            self.logger.debug("Listener started")
            while True:
                message = await self.emit(Commands.RECEIVE_WEBSOCKET_MESSAGE)
                if not message:
                    continue
                asyncio.create_task(self.handle_responses(message))
                self.logger.debug(f"Received a message and sent it to handle_responses.")
                
        except asyncio.CancelledError:
            self.logger.debug("Listener Cancelled.")

    async def stop_listener(self):
        if isinstance(self.listener, asyncio.Task) and not self.listener.done():
            self.listener.cancel()

    async def start_listener(self):
        await self.stop_listener()
        self.listener = asyncio.create_task(self._listener())

    async def handle_responses(self, message):
        response_json = message
        response_code = response_json["op"]
        response_data = response_json.get("d", None)

        if response_code == 1: # Send heartbeat NOW!
            await self.emit(Commands.SEND_ONE_HEARTBEAT)
    
        elif response_code == 10: # HELLO
            self.session.heartbeat_interval = response_data["heartbeat_interval"] / 1000
            await self.emit(Commands.START_HB_LOOP, fire_and_forget=True)
            await self.emit(Commands.START_HB_WATCHDOG, fire_and_forget=True)
            await self.emit(Commands.IDENTIFY_NEEDED)

        elif response_code == 9: # Invalid Session
            await self.emit(Commands.CLIENT_RESET_NEEDED)

        elif response_code == 11:
            await self.emit(Commands.RESTART_HB_WATCHDOG)

        elif response_code == 0:
            await self.emit(Commands.HANDLE_EVENT, fire_and_forget=True, response_json=response_json)


import asyncio
import json
from betterbot import Setup, Events, Intents

# Load bot token from configuration
with open("config.json") as f:
    config = json.load(f)

TOKEN = config["TOKEN"]

class Bot:
    def __init__(self):
        self.bot = Setup(TOKEN, intents=Intents.all)
        self.bot.debug = True  # Enable detailed debug logs
        self.rest = self.bot.rest_client
        self.bot_username = None
        self.event = self.bot.event

    async def main(self):
        # Register event handlers
        await self.event.register(Events.BOT_READY, self.on_ready) # You can register multiple handlers per event.
        await self.event.register(Events.MESSAGE_CREATE, self.on_message)
        await self.bot.start()

    # -----------------------
    # Event Handlers
    # -----------------------
    async def on_ready(self, username, es):
        self.bot_username = username
        print(f"Logged in as {username}")

        # Create a guild slash command
        await es.create_guild_command(
            self.test_command,
            guild_id=GUILD_ID,
            name="testrest",
            description="Test all REST endpoints with defer"
        )

    async def on_message(self, username, content, channel_id):
        if username == self.bot_username:
            return
        print(f"{username}: {content}")
        if content.lower() == "test rest":
            await self.test_rest_methods()

    # -----------------------
    # Slash Command Handler
    # -----------------------
    async def test_command(self, es): # Es stands for eventsend and is the prefered way to respond to events
        # Immediately respond to the interaction
        await es.respond_message(content="Hello, it is starting!")
        await self.test_rest_methods()

    async def test_rest_methods(self):
        # -----------------------
        # Channel Methods
        # -----------------------
        await self.rest.send_message(CHANNEL_ID, "Hello from Betterbot REST test!")

        messages = await self.rest.get_messages(CHANNEL_ID)
        if messages:
            msg_id = messages[0]["id"]
            await self.rest.typing(CHANNEL_ID)
            await self.rest.add_reaction(CHANNEL_ID, msg_id, "👍")
            await self.rest.remove_reaction(CHANNEL_ID, msg_id, "👍")
            await self.rest.pin_message(CHANNEL_ID, msg_id)
            await self.rest.unpin_message(CHANNEL_ID, msg_id)

        # -----------------------
        # Guild Methods
        # -----------------------
        guild_info = await self.rest.get_guild_info(GUILD_ID)
        members = await self.rest.get_guild_members(GUILD_ID, limit=5)
        roles = await self.rest.get_guild_roles(GUILD_ID)

        print("REST example complete!")

# -----------------------
# Run Bot
# -----------------------
asyncio.run(Bot().main())


import httpx
import asyncio
import logging
import time
import json
from socket import gaierror
from . import error
from ..enums import ResponseTypes

class HttpWrapper:
    def __init__(self, token):
        self.logger = logging.getLogger(__name__)
        self.token = token
        self.global_rate_limit = asyncio.Event()
        self.global_rate_limit.set()
        self.http = httpx.AsyncClient(timeout=20)
        self.global_lock = asyncio.Lock()
        self.write_lock = asyncio.Lock()
        self.cache_count = 0
        self.buckets = {}
        self.identifiers = {}

        self._load_cache()

    def _load_cache(self):
        amount = 604800 # One week

        try:
            with open(".__betterbot_cache__.json", "r") as f:
                cache = json.load(f)

                if time.time() - cache["time"] > amount:
                    self.logger.info("Cache is too old. A new cache file will be created.")
                else:
                    for bucket_id, bucket in cache["buckets"].items():
                        bucket["x-ratelimit-limit"] = -1
                        bucket["x-ratelimit-remaining"] = -1
                        bucket["x-ratelimit-reset"] = 0

                        bucket["lock"] = asyncio.Lock()

                    self.buckets = cache["buckets"]
                    self.identifiers = cache["identifiers"]

        except FileNotFoundError:
            self.logger.info("No cache found")
        except json.JSONDecodeError:
            self.logger.error("Cache file invalid.")
        except PermissionError:
            self.logger.error("Permission ERROR: Cannot read cache")
        except KeyError:
            self.logger.error("Cache content not expected.")

            self.identifiers = None
            self.buckets = None

    async def _status_codes(self, response):
        status_code = response.status_code
        self.logger.debug(f"The status code response is {status_code}")

        if status_code == 401:
            raise error.InvalidTokenError("Invalid token!")

        elif status_code in [200, 201, 202, 203, 204, 304]:
            return ResponseTypes.SUCCESS
    
        elif status_code in [403, 404, 400, 405]:
            return ResponseTypes.FAILED
        
        elif status_code in [500, 502, 504]:
            return ResponseTypes.RETRY_DELAY
        
        elif status_code == 429:
            info = response.json()

            sleep = info.get("retry_after", 1)
            global_limit = info.get("global", False)

            if global_limit is True:
                self.logger.warning(f"GLOBAL LIMIT REACHED! Sleeping for {sleep}")
                self.global_rate_limit.clear()
                await asyncio.sleep(sleep)
                self.global_rate_limit.set()
            else:
                self.logger.info(f"Discord rate-limiting. Sleeping for: {sleep}")
                
                await asyncio.sleep(sleep)

            return ResponseTypes.RETRY_NOW
        
        else:
            self.logger.critical("Unhandled Status code treating as unsuccessful.")
            return ResponseTypes.FAILED

    async def _handle_bucket(self, identifier, response):
        self.cache_count += 1
        bucket = response.headers.get("x-ratelimit-bucket", None) 
        limit = int(response.headers.get("x-ratelimit-limit", -1))
        remaining = int(response.headers.get("x-ratelimit-remaining", -1))
        reset = float(response.headers.get("x-ratelimit-reset", 0))
        
        self.logger.debug(f"Bucket response from discord is {bucket}")
        self.logger.debug(f"x-rate-limit response from discord is {limit}")
        self.logger.debug(f"remaining response from discord is {remaining}")
        self.logger.debug(f"Reset unix time response from discord is {reset} and current time in unix is {time.time()}")

        if bucket in self.buckets and identifier not in self.identifiers:
            self.logger.debug(f"New identifier for existing bucket: {bucket}. The new identifier is {identifier}")
            self.identifiers[identifier] = bucket

        elif identifier in self.identifiers and bucket not in self.identifiers.values():
            async with self.write_lock:
                if identifier in self.identifiers and bucket not in self.identifiers.values():
                    self.logger.debug(f"Discord updated their bucket for {identifier} new_bucket: {bucket}")
                    del self.identifiers[identifier]
            
        if bucket is not None and bucket not in self.buckets and identifier not in self.identifiers:
            self.logger.debug(f"Saving a bucket as {bucket}")
            self.buckets[bucket] = {"lock": asyncio.Lock()}
            self.identifiers[identifier] = bucket

        if self.buckets.get(bucket):
            self.logger.debug("New x-rate-limit from discord updating existing.")
            to_update = self.buckets[bucket]
            to_update.update({
                "x-ratelimit-limit": limit,
                "x-ratelimit-remaining": remaining, 
                "x-ratelimit-reset": reset
                })

        if self.cache_count == 25: # Caches every 25 times.
            async with self.write_lock:
                if self.cache_count == 25: # Caches every 25 times.
                    self.cache_count = 0
                    with open(".__betterbot_cache__.json", "w") as f:
                        write_buckets = {key: {k: v for k,v in value.items() if k != "lock"}
                        for key, value in self.buckets.items()
                        }

                        to_cache = {"buckets": write_buckets, "identifiers": self.identifiers, "time": time.time()}
                        json.dump(to_cache, f, indent=2)

    async def _request(self, method, endpoint, identifier, json=None):
        connection = 0
        while True:
            try:
                if json:
                    self.logger.info(f"A json is sent the json is: {json}")

                await self.global_rate_limit.wait()

                response = await self.http.request(method, endpoint, headers={"Authorization": f"Bot {self.token}"}, json=json)

                content_type = response.headers.get("Content-Type", None)

                await self._handle_bucket(identifier, response)
                what_to_do = await self._status_codes(response)

                if what_to_do == ResponseTypes.SUCCESS:
                    self.logger.info("Successfully completed your request")
                    if "application/json" in content_type and response.json() is not None:
                        response_json = response.json()
                        return response_json
                    else:
                        return ResponseTypes.SUCCESS
                                        
                elif what_to_do == ResponseTypes.FAILED:
                    self.logger.warning(f"Your attempt to send a request failed. Code {response.status_code}")
                    return ResponseTypes.FAILED
                
                elif what_to_do == ResponseTypes.RETRY_DELAY:
                    await asyncio.sleep(connection*3)
                    connection += 1
                    if connection == 5:
                        return ResponseTypes.CONNECTION_FAILED
                    continue

                elif what_to_do == ResponseTypes.RETRY_NOW:
                    continue

            except (OSError,httpx.ConnectError, gaierror, httpx.ReadTimeout):
                    self.logger.error(f"Your attempt to send a request failed. Is the network or server down?")
                    return ResponseTypes.CONNECTION_FAILED

    async def request(self, method, endpoint, identifier, json=None):    
        async def send_identifier():
            bucket = self.identifiers[identifier]
            bucket_in_buckets = self.buckets[bucket]
            lock = bucket_in_buckets["lock"]

            async with lock:
                reset_time = bucket_in_buckets["x-ratelimit-reset"]
                current_time = time.time()

                self.logger.debug(f"Existing identifier for this bucket: {bucket}")

                self.logger.debug(f"From self.request() Current unix time is {current_time} and reset time is {reset_time}")

                if bucket_in_buckets["x-ratelimit-remaining"] == 0 and reset_time - current_time > 0:
                    sleep_time = reset_time - current_time
                    self.logger.warning(f"Throttling requests to send as discord requests it. Sleeping for the requested amount: {sleep_time}")
                    
                    await asyncio.sleep(sleep_time)
                            
                return await self._request(method, endpoint, identifier, json)

        if identifier in self.identifiers:
            return await send_identifier()
        
        else:
            await self.global_lock.acquire()
            if identifier in self.identifiers:
                self.global_lock.release()
                return await send_identifier()
            
            response = await self._request(method, endpoint, identifier, json)
            self.global_lock.release()
            return response


from .events import Events
import logging

class HandleEventResponses:
    def __init__(self, eventbus, session):
        self.session = session
        self.emit = eventbus.emit
        self.logger = logging.getLogger(__name__)

        self.dispatch_map = {
            # Lifecycle
            Events.BOT_READY: lambda d: {
                "user_id": d.get("user", {}).get("id"),
                "username": d.get("user", {}).get("username"),
                "guilds": d.get("guilds"),
                "session_id": d.get("session_id"),
                "application_id": d.get("application", {}).get("id"),
                "payload": d,  # full raw payload
            },

            # Guild
            Events.GUILD_CREATE: lambda d: {
                "guild_id": d.get("id"),
                "name": d.get("name"),
                "payload": d
            },
            Events.GUILD_UPDATE: lambda d: {
                "guild_id": d.get("id"),
                "name": d.get("name"),
                "payload": d
            },
            Events.GUILD_DELETE: lambda d: {
                "guild_id": d.get("id"),
                "unavailable": d.get("unavailable"),
                "payload": d
            },

            # Member
            Events.GUILD_MEMBER_ADD: lambda d: {
                "guild_id": d.get("guild_id"),
                "user_id": d.get("user", {}).get("id"),
                "username": d.get("user", {}).get("username"),
                "payload": d
            },
            Events.GUILD_MEMBER_REMOVE: lambda d: {
                "guild_id": d.get("guild_id"),
                "user_id": d.get("user", {}).get("id"),
                "username": d.get("user", {}).get("username"),
                "payload": d
            },
            Events.GUILD_MEMBER_UPDATE: lambda d: {
                "guild_id": d.get("guild_id"),
                "user_id": d.get("user", {}).get("id"),
                "username": d.get("user", {}).get("username"),
                "payload": d
            },

            # Message
            Events.MESSAGE_CREATE: lambda d: {
                "message_id": d.get("id"),
                "channel_id": d.get("channel_id"),
                "guild_id": d.get("guild_id"),
                "user_id": d.get("author", {}).get("id"),
                "username": d.get("author", {}).get("username"),
                "content": d.get("content"),
                "embeds": d.get("embeds", []),
                "payload": d,
            },
            Events.MESSAGE_UPDATE: lambda d: {
                "message_id": d.get("id"),
                "channel_id": d.get("channel_id"),
                "guild_id": d.get("guild_id"),
                "content": d.get("content"),
                "payload": d,
            },
            Events.MESSAGE_DELETE: lambda d: {
                "message_id": d.get("id"),
                "channel_id": d.get("channel_id"),
                "guild_id": d.get("guild_id"),
                "payload": d,
            },
            Events.MESSAGE_DELETE_BULK: lambda d: {
                "message_ids": d.get("ids"),
                "channel_id": d.get("channel_id"),
                "guild_id": d.get("guild_id"),
                "payload": d,
            },

            # Reaction
            Events.MESSAGE_REACTION_ADD: lambda d: {
                "user_id": d.get("user_id"),
                "channel_id": d.get("channel_id"),
                "message_id": d.get("message_id"),
                "guild_id": d.get("guild_id"),
                "emoji": d.get("emoji"),
                "payload": d
            },
            Events.MESSAGE_REACTION_REMOVE: lambda d: {
                "user_id": d.get("user_id"),
                "channel_id": d.get("channel_id"),
                "message_id": d.get("message_id"),
                "guild_id": d.get("guild_id"),
                "emoji": d.get("emoji"),
                "payload": d
            },

            # Channel
            Events.CHANNEL_CREATE: lambda d: {
                "channel_id": d.get("id"),
                "guild_id": d.get("guild_id"),
                "name": d.get("name"),
                "payload": d
            },
            Events.CHANNEL_UPDATE: lambda d: {
                "channel_id": d.get("id"),
                "guild_id": d.get("guild_id"),
                "name": d.get("name"),
                "payload": d
            },
            Events.CHANNEL_DELETE: lambda d: {
                "channel_id": d.get("id"),
                "guild_id": d.get("guild_id"),
                "payload": d
            },

            Events.CONVERSATION_SUMMARY_UPDATE: lambda d: {
                "channel_id": d.get("id"),
                "guild_id": d.get("guild_id"),
                "text": d.get("summary", []).get("text"),
                "last_message_id": d.get("summary", []).get("last_message_id"),
                "message_count": d.get("summary", []).get("message_count"),
                "updated_at": d.get("summary", []).get("updated_at"),
                "payload": d
            },

            # Typing
            Events.TYPING_START: lambda d: {
                "channel_id": d.get("channel_id"),
                "guild_id": d.get("guild_id"),
                "user_id": d.get("user_id"),
                "timestamp": d.get("timestamp"),
                "payload": d
            },

            # Presence
            Events.PRESENCE_UPDATE: lambda d: {
                "user_id": d["user"].get("id"),
                "status": d.get("status"),
                "activities": d.get("activities"),
                "guild_id": d.get("guild_id"),
                "payload": d
            },

            # Voice
            Events.VOICE_STATE_UPDATE: lambda d: {
                "guild_id": d.get("guild_id"),
                "channel_id": d.get("channel_id"),
                "user_id": d.get("user_id"),
                "session_id": d.get("session_id"),
                "payload": d
            },
            Events.VOICE_SERVER_UPDATE: lambda d: {
                "guild_id": d.get("guild_id"),
                "endpoint": d.get("endpoint"),
                "token": d.get("token"),
                "payload": d
            },

            # Interaction
            Events.INTERACTION_CREATE: lambda d: {
                "interaction_id": d.get("id"),
                "interaction_token": d.get("token"),
                "cmd_type": d.get("type"),
                "guild_id": d.get("guild_id"),
                "channel_id": d.get("channel_id"),
                "user": d.get("member", {}).get("user"),
                "user_id": d.get("member", {}).get("user", {}).get("id"),
                "username": d.get("member", {}).get("user", {}).get("username"),
                "options": d.get("data", {}).get("options"),
                "values": [value.get("value") for value in d.get("data", {}).get("options", [])],
                "roles": d.get("member", []).get("roles", []),
                "payload": d,
            },

            # Invite
            Events.INVITE_CREATE: lambda d: {
                "code": d.get("code"),
                "channel_id": d.get("channel_id"),
                "guild_id": d.get("guild_id"),
                "inviter": d.get("inviter"),
                "payload": d
            },
            Events.INVITE_DELETE: lambda d: {
                "code": d.get("code"),
                "channel_id": d.get("channel_id"),
                "guild_id": d.get("guild_id"),
                "payload": d
            },

            # Webhooks
            Events.WEBHOOKS_UPDATE: lambda d: {
                "channel_id": d.get("channel_id"),
                "guild_id": d.get("guild_id"),
                "payload": d
            },
        }

    async def handle_events(self, response_json):
        response_event = response_json["t"]
        response_data = response_json["d"]
        sequence = response_json.get("t", None)
            
        self.logger.debug(f"Debug: Event sent by discord is {response_event}")

        if sequence is not None:
            self.session.sequence = sequence

        if response_event in self.dispatch_map and response_event != Events.INTERACTION_CREATE:
            payload = self.dispatch_map[response_event](response_data)
            await self.emit(response_event, **payload)
            
        elif response_event == Events.INTERACTION_CREATE:
            data = response_data["data"]
            if response_data.get("guild_id"):
                event = f"{Events.INTERACTION_CREATE}:{data["name"]}:{response_data["guild_id"]}"
            else:
                event = f"{Events.INTERACTION_CREATE}:{data["name"]}"

            payload = self.dispatch_map[response_event](response_data)
            
            await self.emit(event, **payload)
            await self.emit(response_event, **payload)
        else:
            self.logger.info(f"Event sent by discord is undefined {response_event}")


class Intents:
    default = 53575421
    all = 53608447

class CommandTypes:
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9 # Can be user or role
    NUMBER = 10
    ATTACHMENT = 11

class ResponseTypes:
    RETRY_NOW = "RETRY_NOW"
    RETRY_DELAY = "RETRY_WITH_DELAY"
    SUCCESS = "SUCCESS"
    FAILED = "ATTEMPT_FAILED"
    CONNECTION_FAILED = "FAILED_DUE_TO_CONNECTION_ISSUES"


class Events:
    EVERYTHING = "listen_to_everything"  # You can subscribe to this if you want to listen to all events.

    # Lifecycle
    BOT_READY = "READY"  # Bot is ready

    # Guild
    GUILD_CREATE = "GUILD_CREATE"  # Guild added
    GUILD_UPDATE = "GUILD_UPDATE"  # Guild updated
    GUILD_DELETE = "GUILD_DELETE"  # Guild removed
    GUILD_ROLE_CREATE = "GUILD_ROLE_CREATE"  # Role created
    GUILD_ROLE_UPDATE = "GUILD_ROLE_UPDATE"  # Role updated
    GUILD_ROLE_DELETE = "GUILD_ROLE_DELETE"  # Role deleted
    GUILD_BAN_ADD = "GUILD_BAN_ADD"  # User banned
    GUILD_BAN_REMOVE = "GUILD_BAN_REMOVE"  # User unbanned
    GUILD_EMOJIS_UPDATE = "GUILD_EMOJIS_UPDATE"  # Emojis updated
    GUILD_STICKERS_UPDATE = "GUILD_STICKERS_UPDATE"  # Stickers updated
    GUILD_INTEGRATIONS_UPDATE = "GUILD_INTEGRATIONS_UPDATE"  # Integrations updated
    GUILD_SCHEDULED_EVENT_CREATE = "GUILD_SCHEDULED_EVENT_CREATE"  # Event created
    GUILD_SCHEDULED_EVENT_UPDATE = "GUILD_SCHEDULED_EVENT_UPDATE"  # Event updated
    GUILD_SCHEDULED_EVENT_DELETE = "GUILD_SCHEDULED_EVENT_DELETE"  # Event deleted

    # Channel
    CHANNEL_CREATE = "CHANNEL_CREATE"  # Channel created
    CHANNEL_UPDATE = "CHANNEL_UPDATE"  # Channel updated
    CHANNEL_DELETE = "CHANNEL_DELETE"  # Channel deleted
    CHANNEL_PINS_UPDATE = "CHANNEL_PINS_UPDATE"  # Pins updated

    # Thread
    THREAD_CREATE = "THREAD_CREATE"  # Thread created
    THREAD_UPDATE = "THREAD_UPDATE"  # Thread updated
    THREAD_DELETE = "THREAD_DELETE"  # Thread deleted
    THREAD_LIST_SYNC = "THREAD_LIST_SYNC"  # Thread list synced
    THREAD_MEMBER_UPDATE = "THREAD_MEMBER_UPDATE"  # Thread member updated
    THREAD_MEMBERS_UPDATE = "THREAD_MEMBERS_UPDATE"  # Thread members updated

    # Member
    GUILD_MEMBER_ADD = "GUILD_MEMBER_ADD"  # Member added
    GUILD_MEMBER_REMOVE = "GUILD_MEMBER_REMOVE"  # Member removed
    GUILD_MEMBER_UPDATE = "GUILD_MEMBER_UPDATE"  # Member updated
    
    # Message
    MESSAGE_CREATE = "MESSAGE_CREATE"  # Message created
    MESSAGE_UPDATE = "MESSAGE_UPDATE"  # Message updated
    MESSAGE_DELETE = "MESSAGE_DELETE"  # Message deleted
    MESSAGE_DELETE_BULK = "MESSAGE_DELETE_BULK"  # Messages bulk deleted

    # Reaction
    MESSAGE_REACTION_ADD = "MESSAGE_REACTION_ADD"  # Reaction added
    MESSAGE_REACTION_REMOVE = "MESSAGE_REACTION_REMOVE"  # Reaction removed
    MESSAGE_REACTION_REMOVE_ALL = "MESSAGE_REACTION_REMOVE_ALL"  # All reactions removed
    MESSAGE_REACTION_REMOVE_EMOJI = "MESSAGE_REACTION_REMOVE_EMOJI"  # Emoji reactions removed

    # Presence / Typing / User
    PRESENCE_UPDATE = "PRESENCE_UPDATE"  # Presence changed
    TYPING_START = "TYPING_START"  # Typing started
    USER_UPDATE = "USER_UPDATE"  # User updated

    # Voice
    VOICE_STATE_UPDATE = "VOICE_STATE_UPDATE"  # Voice state updated
    VOICE_SERVER_UPDATE = "VOICE_SERVER_UPDATE"  # Voice server updated

    # Interaction / Application
    INTERACTION_CREATE = "INTERACTION_CREATE"  # Interaction created
    APPLICATION_COMMAND_CREATE = "APPLICATION_COMMAND_CREATE"  # Command created
    APPLICATION_COMMAND_UPDATE = "APPLICATION_COMMAND_UPDATE"  # Command updated
    APPLICATION_COMMAND_DELETE = "APPLICATION_COMMAND_DELETE"  # Command deleted
    APPLICATION_COMMAND_PERMISSIONS_UPDATE = "APPLICATION_COMMAND_PERMISSIONS_UPDATE"  # Command permissions updated

    # Invite
    INVITE_CREATE = "INVITE_CREATE"  # Invite created
    INVITE_DELETE = "INVITE_DELETE"  # Invite deleted

    # Auto-Moderation
    AUTO_MODERATION_RULE_CREATE = "AUTO_MODERATION_RULE_CREATE"  # Rule created
    AUTO_MODERATION_RULE_UPDATE = "AUTO_MODERATION_RULE_UPDATE"  # Rule updated
    AUTO_MODERATION_RULE_DELETE = "AUTO_MODERATION_RULE_DELETE"  # Rule deleted
    AUTO_MODERATION_ACTION_EXECUTION = "AUTO_MODERATION_ACTION_EXECUTION"  # Action executed
    AUTO_MODERATION_INCIDENT_CREATE = "AUTO_MODERATION_INCIDENT_CREATE"  # Incident created
    AUTO_MODERATION_INCIDENT_UPDATE = "AUTO_MODERATION_INCIDENT_UPDATE"  # Incident updated
    AUTO_MODERATION_INCIDENT_DELETE = "AUTO_MODERATION_INCIDENT_DELETE"  # Incident deleted
    AUTO_MODERATION_MENTION_RAID_DETECTION = "AUTO_MODERATION_MENTION_RAID_DETECTION"  # Raid detected

    # Stage
    STAGE_INSTANCE_CREATE = "STAGE_INSTANCE_CREATE"  # Stage created
    STAGE_INSTANCE_DELETE = "STAGE_INSTANCE_DELETE"  # Stage deleted
    STAGE_INSTANCE_UPDATE = "STAGE_INSTANCE_UPDATE"  # Stage updated

    # Integration
    INTEGRATION_CREATE = "INTEGRATION_CREATE"  # Integration created
    INTEGRATION_UPDATE = "INTEGRATION_UPDATE"  # Integration updated
    INTEGRATION_DELETE = "INTEGRATION_DELETE"  # Integration deleted

    # Webhook
    WEBHOOKS_UPDATE = "WEBHOOKS_UPDATE"  # Webhooks updated

    # Scheduled Event User
    GUILD_SCHEDULED_EVENT_USER_ADD = "GUILD_SCHEDULED_EVENT_USER_ADD"  # User added to event
    GUILD_SCHEDULED_EVENT_USER_REMOVE = "GUILD_SCHEDULED_EVENT_USER_REMOVE"  # User removed from event

    SESSIONS_REPLACE = "SESSIONS_REPLACE" # Session replaced

    CONVERSATION_SUMMARY_UPDATE = "CONVERSATION_SUMMARY_UPDATE" # AI chat summary


from .endpoints import Endpoints

class Channel:
    def __init__(self, wrapper):
        self.send = wrapper.request

    async def get_channel_info(self, channel_id):
        endpoint = Endpoints.CHANNEL_INFO(d={"channel_id": channel_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"get_channel_info:{channel_id}")

    async def send_message(self, channel_id, content=None, embed=None, components=None):
        if not content and not embed and not components:
            raise ValueError("At least one of content, embed, or components must be provided")
        payload = {}
        if content: payload["content"] = content
        if embed: payload["embeds"] = embed
        if components: payload["components"] = components
        endpoint = Endpoints.CHANNEL_MESSAGES(d={"channel_id": channel_id})
        return await self.send(method="post", endpoint=endpoint, identifier=f"send_message:{channel_id}", json=payload)

    async def get_messages(self, channel_id, limit=100, before_id=None):
        endpoint = Endpoints.CHANNEL_MESSAGES(d={"channel_id": channel_id}) + (f"?limit={limit}" if before_id is None else f"?limit={limit}&before={before_id}")
        return await self.send(method="get", endpoint=endpoint, identifier=f"get_messages:{channel_id}")

    async def get_message(self, channel_id, message_id):
        endpoint = Endpoints.CHANNEL_MESSAGE(d={"channel_id": channel_id, "message_id": message_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"get_message:{channel_id}")

    async def delete_message(self, channel_id, message_id):
        endpoint = Endpoints.CHANNEL_MESSAGE(d={"channel_id": channel_id, "message_id": message_id})
        return await self.send(method="delete", endpoint=endpoint, identifier=f"delete_message:{channel_id}")

    async def add_reaction(self, channel_id, message_id, emoji, user_id=None):
        if user_id: endpoint = Endpoints.CHANNEL_REACTION_USER(d={"channel_id": channel_id, "message_id": message_id, "emoji": emoji, "user_id": user_id})
        else: endpoint = Endpoints.CHANNEL_REACTIONS(d={"channel_id": channel_id, "message_id": message_id, "emoji": emoji})
        return await self.send(method="put", endpoint=endpoint, identifier=f"add_reaction:{channel_id}")

    async def remove_reaction(self, channel_id, message_id, emoji, user_id=None):
        if user_id: endpoint = Endpoints.CHANNEL_REACTION_USER(d={"channel_id": channel_id, "message_id": message_id, "emoji": emoji, "user_id": user_id})
        else: endpoint = Endpoints.CHANNEL_REACTIONS(d={"channel_id": channel_id, "message_id": message_id, "emoji": emoji})
        return await self.send(method="delete", endpoint=endpoint, identifier=f"remove_reaction:{channel_id}")

    async def bulk_delete_messages(self, channel_id, message_ids):
        payload = {"messages": message_ids}
        endpoint = Endpoints.CHANNEL_BULK_DELETE_MESSAGES(d={"channel_id": channel_id})
        return await self.send(method="post", endpoint=endpoint, identifier=f"bulk_delete:{channel_id}", json=payload)

    async def get_pins(self, channel_id):
        endpoint = Endpoints.CHANNEL_PINS(d={"channel_id": channel_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"get_pins:{channel_id}")

    async def pin_message(self, channel_id, message_id):
        endpoint = Endpoints.CHANNEL_PIN(d={"channel_id": channel_id, "message_id": message_id})
        return await self.send(method="put", endpoint=endpoint, identifier=f"pin_message:{channel_id}")

    async def unpin_message(self, channel_id, message_id):
        endpoint = Endpoints.CHANNEL_PIN(d={"channel_id": channel_id, "message_id": message_id})
        return await self.send(method="delete", endpoint=endpoint, identifier=f"unpin_message:{channel_id}")

    async def typing(self, channel_id):
        endpoint = Endpoints.CHANNEL_TYPING(d={"channel_id": channel_id})
        return await self.send(method="post", endpoint=endpoint, identifier=f"typing:{channel_id}")


from ..command_related.commands import Commands
import logging

class Identify:
    def __init__(self, commandbus, session):
        self.logger = logging.getLogger(__name__)
        self.emit = commandbus.emit
        self.token = session.token
        self.intents = session.intents

    async def identify(self):
        payload = {
            "op": 2,
            "d": {
            "token": self.token,
            "intents": self.intents,
            "properties": {
                "os": "Windows",
                "browser": "Betterbot",
                "device": "Betterbot",
                },
            }
        }

        self.logger.debug("Identifying...")
        await self.emit(Commands.SEND_WEBSOCKET_MESSAGE, fire_and_forget=False, data=payload)


from .endpoints import Endpoints
import logging

class InteractionTypes:
    DEFERRED_CHANNEL_MESSAGE = 2
    DEFERRED_UPDATE_MESSAGE = 3
    RESPOND_WITH_MESSAGE = 4
    UPDATE_MESSAGE = 5
    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 6
    MODAL = 7

class Interaction:
    def __init__(self, wrapper):
        self.send = wrapper.request

    async def send_interaction_callback(self, interaction_id, interaction_token, cmd_type, content=None, embed=None, components=None, ephemeral=False):
        payload = {}
        data = {}
        payload["type"] = cmd_type
        if content: data["content"] = content
        if embed: data["embeds"] = embed
        if components: data["components"] = components
        if ephemeral: data["flags"] = 64
        if data: payload["data"] = data
        endpoint = Endpoints.INTERACTION_CALLBACK(d={"interaction_id": interaction_id, "interaction_token": interaction_token})
        return await self.send(method="post", endpoint=endpoint, identifier=f"interaction_callback", json=payload)

    async def send_interaction_followup(self, application_id, interaction_token, content=None, embed=None, components=None, ephemeral=False):
        payload = {}
        data = {}
        if content: data["content"] = content
        if embed: data["embeds"] = embed
        if components: data["components"] = components
        if ephemeral: data["flags"] = 64
        if data: payload["data"] = data
        endpoint = Endpoints.INTERACTION_FOLLOWUP(d={"application_id": application_id, "interaction_token": interaction_token})
        return await self.send(method="post", endpoint=endpoint, identifier=f"interaction_followup:{application_id}", json=payload)

    async def get_original_interaction_message(self, application_id, interaction_token):
        endpoint = Endpoints.INTERACTION_ORIGINAL_MESSAGE(d={"application_id": application_id, "interaction_token": interaction_token})
        return await self.send(method="get", endpoint=endpoint, identifier=f"interaction_original_message:{application_id}")

    async def update_interaction_message(self, application_id, interaction_token, content=None, embed=None, components=None, message_id=None):
        payload = {}
        data = {}
        if content: data["content"] = content
        if embed: data["embeds"] = embed
        if components: data["components"] = components
        if data: payload["data"] = data
        endpoint = Endpoints.INTERACTION_UPDATE_MESSAGE(d={"application_id": application_id, "interaction_token": interaction_token, "message_id": message_id or "@original"})
        return await self.send(method="patch", endpoint=endpoint, identifier=f"interaction_update_message:{application_id}", json=payload)


import websockets
import logging
import json
import asyncio
from .. import error
from ..command_related.commands import Commands

class WebsocketHandler:
    gateway_url = "wss://gateway.discord.gg/?v=10&encoding=json"

    def __init__(self, commandbus):
        self.logger = logging.getLogger(__name__)
        self.emit = commandbus.emit
        self.reset_running = False
        self.ws = None

    async def reset(self):
        if self.reset_running is True:
            return "BUSY"
        
        self.reset_running = True

        error_handling = error.HandleGatewayErrors()

        while True:
            try:
                await self.emit(Commands.STOP_LISTENER)
                await self.emit(Commands.STOP_HB_WATCHDOG)
                await self.emit(Commands.STOP_HB_LOOP)

                self.logger.debug("Attempting to connect to websocket.")

                self.ws = await websockets.connect(self.gateway_url, ping_interval=None, ping_timeout=None, open_timeout=5)

                self.logger.debug(f"Websocket connected successfully.")

                self.reset_running = False
                await self.emit(Commands.START_LISTENER, fire_and_forget=True)
                return
            
            except Exception as e:
                self.logger.warning("Reset attempt failed trying again.")
                await error_handling.handle_gateway_errors(self.ws, e)
                continue

    async def send_websocket_message(self, data):
        self.logger.debug(f"Sending websocket message {data}")
        try:
            await self.ws.send(json.dumps(data))

        except Exception as e:
            self.logger.error(f"Connection closed while trying to send a message. Code {self.ws.close_code}. Attempting to restart. {e}")
            await self.emit(Commands.CLIENT_RESET_NEEDED, fire_and_forget=True)
            await asyncio.Future() # Will be cancelled shortly

    async def receive_message(self):
        self.logger.debug("Waiting for message...")
        try:
            msg = await self.ws.recv()

        except Exception as e:
            self.logger.error(f"Connection closed while trying to receive message. Code {self.ws.close_code}. Attempting to restart. {e}")
            await self.emit(Commands.CLIENT_RESET_NEEDED, fire_and_forget=True)
            await asyncio.Future() # Will be cancelled shortly

        if type(msg) is str:
            return json.loads(msg)
        
        elif type(msg) is bytes:
            self.logger.critical("Received BYTES from DISCORD. This should not have happened.")
            raise RuntimeError()


from .endpoints import Endpoints

class Guild:
    def __init__(self, wrapper):
        self.send = wrapper.request

    async def get_guild_info(self, guild_id):
        endpoint = Endpoints.GUILD_INFO(d={"guild_id": guild_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"guild_info:{guild_id}")

    async def get_guild_members(self, guild_id, limit=100, after=None):
        endpoint = Endpoints.GUILD_MEMBERS_LIST(d={"guild_id": guild_id}) + (f"?limit={limit}" + (f"&after={after}" if after else ""))
        return await self.send(method="get", endpoint=endpoint, identifier=f"guild_members:{guild_id}")

    async def get_guild_member(self, guild_id, user_id):
        endpoint = Endpoints.GUILD_MEMBER_INFO(d={"guild_id": guild_id, "user_id": user_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"guild_member:{guild_id}")

    async def get_guild_roles(self, guild_id):
        endpoint = Endpoints.GUILD_ROLES_LIST(d={"guild_id": guild_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"guild_roles:{guild_id}")

    async def get_guild_role(self, guild_id, role_id):
        endpoint = Endpoints.GUILD_ROLE_INFO(d={"guild_id": guild_id, "role_id": role_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"guild_role:{guild_id}")

    async def add_member_role(self, guild_id, user_id, role_id):
        endpoint = Endpoints.MEMBER_ROLE_ASSIGN(d={"guild_id": guild_id, "user_id": user_id, "role_id": role_id})
        return await self.send(method="put", endpoint=endpoint, identifier=f"guild_roles_assign:{guild_id}")

    async def remove_member_role(self, guild_id, user_id, role_id):
        endpoint = Endpoints.MEMBER_ROLE_ASSIGN(d={"guild_id": guild_id, "user_id": user_id, "role_id": role_id})
        return await self.send(method="delete", endpoint=endpoint, identifier=f"guild_role_remove:{guild_id}")

    async def member_roles(self, guild, user_id, roles=None):
        payload = {"roles": []}
        if roles: payload["roles"] = roles
        endpoint = Endpoints.MEMBER_ROLE_ASSIGN_JSON(d={"guild_id": guild, "user_id": user_id})
        return await self.send(method="patch", endpoint=endpoint, identifier=f"guild_roles_remove:{guild}", json=payload)

    async def get_guild_bans(self, guild_id):
        endpoint = Endpoints.GUILD_BANS_LIST(d={"guild_id": guild_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"guild_bans:{guild_id}")

    async def get_guild_ban(self, guild_id, user_id):
        endpoint = Endpoints.GUILD_BAN_INFO(d={"guild_id": guild_id, "user_id": user_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"guild_ban:{guild_id}")

    async def prune_guild_members(self, guild_id, days=None, compute_prune_count=False):
        payload = {}
        if days is not None: payload["days"] = days
        payload["compute_prune_count"] = compute_prune_count
        endpoint = Endpoints.GUILD_PRUNE_MEMBERS(d={"guild_id": guild_id})
        return await self.send(method="post", endpoint=endpoint, identifier=f"guild_prune:{guild_id}", json=payload)

    async def get_guild_channels(self, guild_id):
        endpoint = Endpoints.GUILD_CHANNELS_LIST(d={"guild_id": guild_id})
        return await self.send(method="get", endpoint=endpoint, identifier=f"guild_channels:{guild_id}")


import logging
import asyncio
from websockets import ConnectionClosed, InvalidHandshake, ConnectionClosedError
from socket import gaierror
from ssl import SSLError

class InvalidTokenError(Exception): pass
class MaxRetriesAttempted(Exception): pass
class InvalidIntentsError(Exception): pass
class DisallowedIntentsError(Exception): pass

class HandleGatewayErrors:
    def __init__(self):
        self.reset_counter = 0
        self.logger = logging.getLogger(__name__)

    async def handle_gateway_errors(self, ws, e):
        if isinstance(e, (OSError, ConnectionClosedError, ConnectionClosed, gaierror, InvalidHandshake, SSLError)):
            pass
        else:
            raise e
        
        if hasattr(ws, "close_code") and ws.close_code is not None:
            if ws.close_code == 1000:
                self.logger.warning("Discord closed for maintainence or related. Close code 1000.")

            elif ws.close_code == 4001:
                self.logger.error("Please report this. Unknown opcode was sent. Code 4001")

            elif ws.close_code == 4003:
                self.logger.error("Discord responded with NOT_AUTHENTICATED code 4003. This should be impossible. Report this please.")

            elif ws.close_code == 4004:
                raise InvalidTokenError("Token is invalid. Code 4004")

            elif ws.close_code == 4005:
                self.logger.error("Discord responded with ALREADY_AUTHENTICATED. This shouldn't happen. Code 4005")

            elif ws.close_code == 4008:
                self.logger.warning("Discord is ratelimiting gateway connect attempts. Sleeping for ten seconds before retry. Code 4008")
                await asyncio.sleep(10)

            elif ws.close_code == 4013:
                raise InvalidIntentsError("Invalid Intents. Code 4013")

            elif ws.close_code == 4014:
                raise DisallowedIntentsError("You must open discord developer portal and enable intents. Code 4014.")

        if hasattr(ws, "close_code") and ws.close_code is None:
            self.logger.debug("Attempting to close connection")
            await ws.close() 

        self.reset_counter += 1

        if self.reset_counter == 100:
            raise MaxRetriesAttempted("Multiple retries attempted and failed.")
        else:
            await asyncio.sleep(self.reset_counter * 2)


import asyncio
import json
import sqlite3
import time
from betterbot import InteractionTypes
from betterbot import Types

class Ban:
    con = sqlite3.connect("database.db", check_same_thread=False)
    cur = con.cursor()
    cur.execute("CREATE TABLE IF NOT EXISTS banlist(user_id TEXT PRIMARY KEY, roles, duration, time_stamp)")
    lock = asyncio.Lock()

    def __init__(self, mod_roles, rest_client, guild, ban_role):
        self.rest_client = rest_client
        self.mod_roles = mod_roles
        self.guild = guild
        self.ban_role = ban_role

    async def _check_mod(self, roles):
        if any(r in self.mod_roles for r in roles):
            return True
        else:
            return False
        
    async def _respond(self, response, token, interaction_id):
        if response != Types.FAILED and response != Types.CONNECTION_FAILED:
            await self.rest_client.send_interaction_callback(interaction_id, token, InteractionTypes.RESPOND_WITH_MESSAGE, content=f"Successfully completed your requested action.")
            return True
        else:
            await self.rest_client.send_interaction_callback(interaction_id, token, InteractionTypes.RESPOND_WITH_MESSAGE, content=f"Oops something went wrong.")
            return False

    async def ban(self, values, interaction_id, roles, token):
        if await self._check_mod(roles):
            user_to_ban = values[0]
            unit = values[1]
            duration = values[2]
            
            payload = await self.rest_client.get_guild_member(self.guild, user_to_ban)
            check_roles = payload.get("roles")

            if check_roles != self.ban_role:
                response = await self.rest_client.member_roles(self.guild, user_to_ban, roles=self.ban_role)

            all_roles = json.dumps(check_roles)

            if await self._respond(response, token, interaction_id):
                if unit == "days": amount = duration*86400
                if unit == "hours": amount = duration*3600
                if unit == "minutes": amount = duration*60
                
                await self.store_banned(user_to_ban, all_roles, duration=amount)
        else:
            await self.rest_client.send_interaction_callback(interaction_id, token, InteractionTypes.RESPOND_WITH_MESSAGE, content="You do not have permission to use that command!")

    async def store_banned(self, user_id, roles, duration):
        async with Ban.lock:
            current_time = time.time()
            Ban.cur.execute("INSERT OR REPLACE INTO banlist VALUES (?,?,?,?)", (user_id, roles, duration, current_time))
            Ban.con.commit()

    async def unban(self, values, roles, token, interaction_id):
        if await self._check_mod(roles):
            user_to_unban = values[0]
            async with Ban.lock:
                Ban.cur.execute("SELECT roles FROM banlist WHERE user_id = ?", (user_to_unban,))
                fetched = Ban.cur.fetchone()

                if fetched is None:
                    await self.rest_client.send_interaction_callback(interaction_id, token, InteractionTypes.RESPOND_WITH_MESSAGE, content="User is not banned.")
                    return

                user_to_unban_roles = json.loads(fetched[0])
                response = await self.rest_client.member_roles(self.guild, user_to_unban, roles=user_to_unban_roles)
                await self._respond(response, token, interaction_id)
        else:
            await self.rest_client.send_interaction_callback(interaction_id, token, InteractionTypes.RESPOND_WITH_MESSAGE, content="You do not have permission to use that command!")

    async def polling_unban(self):
        while True:
            await asyncio.sleep(60)
            async with Ban.lock:
                Ban.cur.execute("SELECT user_id, roles, duration, time_stamp FROM banlist")
                users = Ban.cur.fetchall()

                for user in users:
                    difference = time.time() - user[3]
                    if difference > user[2]:
                        await self.rest_client.member_roles(self.guild, user[0], roles=json.loads(user[1]))
                        Ban.cur.execute("DELETE FROM banlist WHERE user_id = ?", (user[0],))
                        Ban.con.commit()


from setuptools import setup, find_packages

setup(
    name="betterbot",
    packages=["betterbot", "betterbot/tests","betterbot/gateway_client/command_related", "betterbot/gateway_client", "betterbot/rest_client", "betterbot/rest_client/rest_handlers", "betterbot/gateway_client/connection", "betterbot/gateway_client/event_related", "betterbot/gateway_client/handlers"],  # Explicitly list top-level packages
)


class Endpoints:
    BASE_URL = "https://discord.com/api/v10/"

    # Interactions / Slash Commands
    INTERACTION_CALLBACK = lambda d: f"{Endpoints.BASE_URL}interactions/{d['interaction_id']}/{d['interaction_token']}/callback"
    INTERACTION_FOLLOWUP = lambda d: f"{Endpoints.BASE_URL}webhooks/{d['application_id']}/{d['interaction_token']}"
    INTERACTION_ORIGINAL_MESSAGE = lambda d: f"{Endpoints.BASE_URL}webhooks/{d['application_id']}/{d['interaction_token']}/messages/@original"
    INTERACTION_UPDATE_MESSAGE = lambda d: f"{Endpoints.BASE_URL}webhooks/{d['application_id']}/{d['interaction_token']}/messages/{d.get('message_id', '@original')}"

    # Application Commands
    APPLICATION_GLOBAL_COMMANDS = lambda d: f"{Endpoints.BASE_URL}applications/{d['application_id']}/commands"
    APPLICATION_COMMAND = lambda d: f"{Endpoints.BASE_URL}applications/{d['application_id']}/commands/{d['command_id']}"
    APPLICATION_GUILD_COMMANDS = lambda d: f"{Endpoints.BASE_URL}applications/{d['application_id']}/guilds/{d['guild_id']}/commands"
    APPLICATION_GUILD_COMMAND = lambda d: f"{Endpoints.BASE_URL}applications/{d['application_id']}/guilds/{d['guild_id']}/commands/{d['command_id']}"
    APPLICATION_COMMAND_PERMISSIONS = lambda d: f"{Endpoints.BASE_URL}applications/{d['application_id']}/guilds/{d['guild_id']}/commands/{d['command_id']}/permissions"

    # Channels
    CHANNEL_INFO = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}"
    CHANNEL_MESSAGES = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/messages"
    CHANNEL_MESSAGE = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/messages/{d['message_id']}"
    CHANNEL_REACTIONS = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/messages/{d['message_id']}/reactions/{d['emoji']}/@me"
    CHANNEL_REACTION_USER = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/messages/{d['message_id']}/reactions/{d['emoji']}/{d['user_id']}"
    CHANNEL_BULK_DELETE_MESSAGES = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/messages/bulk-delete"
    CHANNEL_PINS = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/pins"
    CHANNEL_PIN = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/pins/{d['message_id']}"
    CHANNEL_TYPING = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/typing"

    # Threads
    THREAD_CREATE = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/messages/{d['message_id']}/threads"
    THREAD_MEMBERS_LIST = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/thread-members"
    THREAD_MEMBER_INFO = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/thread-members/{d['user_id']}"
    THREAD_ACTIVE_LIST = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/threads/active"

    # Guilds
    GUILD_INFO = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}"
    GUILD_MEMBERS_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/members"
    GUILD_MEMBER_INFO = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/members/{d['user_id']}"
    GUILD_ROLES_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/roles"
    GUILD_ROLE_INFO = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/roles/{d['role_id']}"
    MEMBER_ROLE_ASSIGN = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/members/{d['user_id']}/roles/{d['role_id']}"
    MEMBER_ROLE_ASSIGN_JSON = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/members/{d['user_id']}"
    GUILD_BANS_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/bans"
    GUILD_BAN_INFO = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/bans/{d['user_id']}"
    GUILD_PRUNE_MEMBERS = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/prune"
    GUILD_CHANNELS_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/channels"

    # Emojis
    GUILD_EMOJIS_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/emojis"
    GUILD_EMOJI_INFO = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/emojis/{d['emoji_id']}"

    # Stickers
    GUILD_STICKERS_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/stickers"
    GUILD_STICKER_INFO = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/stickers/{d['sticker_id']}"

    # Invites
    INVITE_INFO = lambda d: f"{Endpoints.BASE_URL}invites/{d['invite_code']}"
    CHANNEL_INVITES_LIST = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/invites"
    GUILD_INVITES_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/invites"

    # Webhooks
    CHANNEL_WEBHOOKS_LIST = lambda d: f"{Endpoints.BASE_URL}channels/{d['channel_id']}/webhooks"
    GUILD_WEBHOOKS_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/webhooks"
    WEBHOOK_INFO = lambda d: f"{Endpoints.BASE_URL}webhooks/{d['webhook_id']}"
    WEBHOOK_INFO_WITH_TOKEN = lambda d: f"{Endpoints.BASE_URL}webhooks/{d['webhook_id']}/{d['webhook_token']}"

    # Users
    USER_ME = lambda d: f"{Endpoints.BASE_URL}users/@me"
    USER_INFO = lambda d: f"{Endpoints.BASE_URL}users/{d['user_id']}"
    USER_ME_GUILDS_LIST = lambda d: f"{Endpoints.BASE_URL}users/@me/guilds"
    USER_ME_GUILD_INFO = lambda d: f"{Endpoints.BASE_URL}users/@me/guilds/{d['guild_id']}"

    # Voice
    VOICE_STATE_INFO = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/voice-states/{d['user_id']}"

    # Scheduled Events
    SCHEDULED_EVENTS_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/scheduled-events"
    SCHEDULED_EVENT_INFO = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/scheduled-events/{d['event_id']}"
    SCHEDULED_EVENT_USERS = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/scheduled-events/{d['event_id']}/users"

    # Stage Instances
    STAGE_INSTANCE_INFO = lambda d: f"{Endpoints.BASE_URL}stage-instances/{d['channel_id']}"

    # Audit Logs
    AUDIT_LOGS_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/audit-logs"

    # Auto Moderation Rules
    AUTO_MOD_RULES_LIST = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/auto-moderation/rules"
    AUTO_MOD_RULE_INFO = lambda d: f"{Endpoints.BASE_URL}guilds/{d['guild_id']}/auto-moderation/rules/{d['rule_id']}"


from .gateway_client.handlers.websocket_handler import WebsocketHandler
from .gateway_client.command_related.commandbus import Commandbus
from .gateway_client import error
from .gateway_client.command_related.commands import Commands
from .gateway_client.connection.identify import Identify
from .gateway_client.connection.heartbeatservice import HeartbeatService
from .gateway_client.connection.session import Session
from .gateway_client.event_related.eventbus import EventBus
from .gateway_client.handlers.handle_responses import HandleResponses
from .gateway_client.event_related.handle_event_responses import HandleEventResponses
from .rest_client.restclient import RESTClient

import sys
import asyncio
import logging

class Setup:
    def __init__(self, token: str, intents: int):
        if token is None:
            raise error.InvalidTokenError("You must provide your discord token.")
        
        if intents is None:
            raise error.InvalidIntentsError("You must provide Intents.")

        self.debug_logs = False
        self.test_mode = False

        self.token = token
        self.intents = intents

        self.rest_client = RESTClient(token, self.event)
        self.event = EventBus(self.rest_client)

        self.stop = asyncio.Event()

    def _start_logging(self):
        sys.stdout.reconfigure(encoding="utf-8")
        logging.basicConfig(
            level=logging.DEBUG if self.debug_logs else logging.WARNING,
            handlers=[
                logging.FileHandler("log.txt", mode="w", encoding="utf-8"),
                logging.StreamHandler(sys.stdout)
            ],
        )

    async def start(self):
        self._start_logging()

        commandbus = Commandbus()
        session = Session()

        session.token = self.token
        session.intents = self.intents

        websocket_handler = WebsocketHandler(commandbus)
        handle_events = HandleEventResponses(self.event, session)
        identify = Identify(commandbus, session)
        heartbeat_service = HeartbeatService(commandbus, session)
        handle_responses = HandleResponses(commandbus, session)


        if self.test_mode is True:
            return commandbus, session, identify, heartbeat_service, handle_responses, websocket_handler, self.event

        await commandbus.register(Commands.IDENTIFY_NEEDED, identify.identify)

        await commandbus.register(Commands.CLIENT_RESET_NEEDED, websocket_handler.reset)
        await commandbus.register(Commands.SEND_WEBSOCKET_MESSAGE, websocket_handler.send_websocket_message)
        await commandbus.register(Commands.RECEIVE_WEBSOCKET_MESSAGE, websocket_handler.receive_message)

        await commandbus.register(Commands.RESTART_HB_WATCHDOG, heartbeat_service.start_hb_watchdog)
        await commandbus.register(Commands.START_HB_WATCHDOG, heartbeat_service.start_hb_watchdog)
        await commandbus.register(Commands.STOP_HB_WATCHDOG, heartbeat_service.stop_hb_watchdog)
        await commandbus.register(Commands.START_HB_LOOP, heartbeat_service.start_hb_loop)
        await commandbus.register(Commands.STOP_HB_LOOP, heartbeat_service.stop_hb_loop)
        await commandbus.register(Commands.SEND_ONE_HEARTBEAT, heartbeat_service.send_heartbeat)

        await commandbus.register(Commands.START_LISTENER, handle_responses.start_listener)
        await commandbus.register(Commands.STOP_LISTENER, handle_responses.stop_listener)
        await commandbus.register(Commands.HANDLE_EVENT, handle_events.handle_events)

        await websocket_handler.reset()
        await self.stop.wait()


class Commands:
    IDENTIFY_NEEDED = "identify_requested"
    SEND_WEBSOCKET_MESSAGE = "send_websocket_message_requested"
    CLIENT_RESET_NEEDED = "client_reset_needed"
    STOP_HB_WATCHDOG = "stop_heartbeat_watchdog"
    START_HB_WATCHDOG = "start_heartbeat_watchdog"
    RESTART_HB_WATCHDOG = "restart_heartbeat_watchdog"
    START_HB_LOOP = "start_heartbeat_loop"
    STOP_HB_LOOP = "stop_heartbeat_loop"
    START_LISTENER = "start_listener"
    STOP_LISTENER = "stop_listener"
    RECEIVE_WEBSOCKET_MESSAGE = "receive_websocket_message"
    SEND_ONE_HEARTBEAT = "send_one_heartbeat"
    HANDLE_EVENT = "handle_response_events"


import asyncio
from .events import Events
from ...rest_client.eventsend import es
import logging
import inspect

class EventBus:
    def __init__(self, rest_client):
        self.logger = logging.getLogger(__name__)
        self.rest_client = rest_client
        self.registry = {}
        self.waiters = {}
        self.tasks = set() # Required to avoid garbage collection.

    async def register(self, event_type, *handlers):
        self.logger.debug(f"Registering {event_type} for {handlers}")
        if event_type not in self.registry:
            self.registry[event_type] = []
        
        self.registry[event_type].extend(handlers)

    async def unregister(self, event_type, handler=None):
        if event_type in self.registry:
            if handler is None:
                self.logger.debug(f"Unregistering all handlers in {event_type}")
                del self.registry[event_type]
            else:
                self.logger.debug(f"Unregistering {handler} for {event_type}")
                if handler in self.registry[event_type]:
                    self.registry[event_type].remove(handler)

    async def emit(self, event_type, **kwargs):
        self.logger.debug(f"Emitting {event_type} with {kwargs}")
        handlers = self.registry.get(event_type, [])
        self.logger.debug(f"Handlers registered for {event_type} is {handlers} and they are now gonna be executed.")

        for handler in self.registry.get(Events.EVERYTHING, []):
            self.logger.debug(f"Everything handler exists.")
            task = asyncio.create_task(handler(event_type, **kwargs))
            self.tasks.add(task)
            task.add_done_callback(self.tasks.discard)
            self.logger.debug(f"Notified everything handler for {event_type}")

        for handler in handlers:
            allowed_params = inspect.signature(handler).parameters.keys()
            filtered_params = {param_name: value for param_name, value in kwargs.items() if param_name in allowed_params}
            
            if "es" in allowed_params:
                filtered_params["es"] = es(self.rest_client, **kwargs)

            self.logger.debug(f"Notifying {handler} for {event_type}")

            task = asyncio.create_task(handler(**filtered_params))
            self.tasks.add(task)
            task.add_done_callback(self.tasks.discard)

        for waiter in self.waiters.get(event_type, []):
            self.logger.debug(f"Notified waiters for {event_type}")
            waiter.set()

    async def wait_for_event(self, event_type):
        if event_type not in self.waiters:
            self.waiters[event_type] = []

        event = asyncio.Event()
        self.waiters[event_type].append(event)

        await event.wait()

        self.waiters[event_type].remove(event)